// CSE 681 lab 3
// lab3.h
// Kyle Moore

#define FAR_PLANE 100000
#define EPSILON 0.001f
#define MAXREFLRECURSION 3
#define MAXREFRRECURSION 3
#define AA_ALPHA 0.1f // difference required to trigger anti-aliasing

#define SAMPLING 2 // 0 = uniform, 1 = jittered, 2 = random
#define RECONST 2 // 0 = box, 1 = tent, 2 = mitchell-netravali

#include <math.h>
#include <stdio.h>
#include <iostream>
#include <ostream>
#include <stdlib.h>
#include <stdlib.h>
#include <time.h>


// OpenInventor include files
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoLight.h>
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>

#include "OSUInventor.h"

using namespace System::Drawing;

// return random float from -1 to 1
float randFloat() {
    return (((float)rand())/RAND_MAX - 0.5f);
}

class Object3d;

class Ray {
public:
	Ray () {}
	Ray (const SbVec3f &dir, const SbVec3f &orig)
	{_dir = dir; _orig = orig;}
	Ray (const Ray& r) {*this=r;}

	const SbVec3f &origin() {return _orig;}
	const SbVec3f &direction() {return _dir;}
	SbVec3f pointAtParameter (float t) {return _orig+t*_dir;}
	void transform(SbMatrix &matr) {
		matr.multDirMatrix(_dir, _dir);
		matr.multVecMatrix(_orig, _orig);
	}

private:
	SbVec3f _dir;
	SbVec3f _orig;
};

class Hit { // object to return all info about the intersection
public:
	Object3d *obj;
	SbVec3f _normal;
	SbVec3f _position;
};

class Object3d {
public:
	Object3d() {}

	virtual bool intersect(Ray &r, Hit &h, float &tmin)=0;
	// return transparency of object if intersect, else 1.0f
	virtual float intersectTrans(Ray &r, float &tmin, float &trans)=0;
	void setAmbient(const SbColor &ambient) {
		_ambient = ambient;
	}
	void setDiffuse(const SbColor &diffuse) {
		_diffuse = diffuse;
	}
	void setSpecular(const SbColor &specular) {
		_specular = specular;
	}
	void setShininess(const float shininess) {
		_shininess = shininess;
	}
	void setTransparency(const float transparency) {
		_transparency = transparency;
	}

	virtual void setTransform(const SbVec3f &t, const SbRotation &r, const SbVec3f &s)=0;

	// attributes
	SbColor _ambient;
	SbColor _diffuse;
	SbColor _specular;
	float _shininess;
	float _transparency;
	int _texture;

	// transformation matrices
	SbMatrix _transform, _transInv;
};

class Sphere : public Object3d {
public:
	Sphere() {
		_radius = 1.f;
		_texture = 0;
	}
	Sphere(const float radius, const int texture) {
		_radius = radius;
		_texture = texture;
	}

	bool intersect(Ray &r, Hit &h, float &tmin) {
		// transform ray into object space
		Ray objRay = Ray(r.direction(), r.origin());
		objRay.transform(_transInv);

		// Geometric approach
		SbVec3f ce = SbVec3f(-1.f*objRay.origin());
		float q = ce.length();
		SbVec3f dir = objRay.direction();
		float factor = dir.normalize();
		float m = ce.dot(dir);

		float s = q*q - m*m;
		float radius = _radius*_radius;
		if(radius < s) // Test for intersection
			return false;

		float k = sqrt(radius - s);
		float t1 = m + k;
		float t2 = m - k;
		if(t1 > 0.f){
			if(t2 > 0.f) {
				// the smaller t is correct
				if(t1 < t2){
					tmin = t1;
				}
				else{
					tmin = t2;
				}
			}
			else {
				// t1 is the correct t
				tmin = t1;
			}
		}
		else if(t2 >= 1){
			// t2 is the correct t
			tmin = t2;
		}
		else
			return false; // neither t is valid

		// fix t
		tmin /= factor;

		// create the hit object
		h._position = objRay.pointAtParameter(tmin);

		// calculate normal
		h._normal = h._position;
		_transform.multVecMatrix(h._position, h._position);
		_transform.multDirMatrix(h._normal, h._normal);
		h._normal.normalize();

		// set object
		h.obj = (Object3d*)this;

		return true;
	}

	float intersectTrans(Ray &r, float &tmin, float &trans) {
		// transform ray into object space
		Ray objRay = Ray(r.direction(), r.origin());
		objRay.transform(_transInv);

		// Geometric approach
		SbVec3f ce = SbVec3f(-1.f*objRay.origin());
		float q = ce.length();
		SbVec3f dir = objRay.direction();
		float factor = dir.normalize();
		float m = ce.dot(dir);

		float s = q*q - m*m;
		float radius = _radius*_radius;
		if(radius < s) // Test for intersection
			return false;

		float k = sqrt(radius - s);
		float t1 = m + k;
		float t2 = m - k;
		if(t1 > 0.f){
			if(t2 > 0.f) {
				// the smaller t is correct
				if(t1 < t2){
					tmin = t1;
				}
				else{
					tmin = t2;
				}
			}
			else {
				// t1 is the correct t
				tmin = t1;
			}
		}
		else if(t2 >= 1){
			// t2 is the correct t
			tmin = t2;
		}
		else
			return false; // neither t is valid

		// fix t
		tmin /= factor;

		trans = _transparency;
		return true;
	}

	void setTransform(const SbVec3f &t, const SbRotation &r, const SbVec3f &s) {
		_transform.setTransform(t, r, s);
		_transInv = _transform.inverse();
	}

private:
	float _radius;
};

class Cube : public Object3d {
public:
	Cube() {
		_width = 1.0f;
		_height = 1.0f;
		_depth = 1.0f;
		_transform = SbMatrix();
		_texture = 0;
	}

	Cube(const float width, const float height, const float depth, const int texture) {
		_width = width;
		_height = height;
		_depth = depth;
		_transform = SbMatrix();
		_texture = texture;
	}

	void setTransform(const SbVec3f &t, const SbRotation &r, const SbVec3f &s) {
		_transform.setTransform(t, r, s);
		_transInv = _transform.inverse();
	}

	bool intersect(Ray &r, Hit &h, float &tmin) {
		// transform ray into object space
		Ray objRay = Ray(r.direction(), r.origin());
		objRay.transform(_transInv);

		float tnear = -FAR_PLANE;
		float tfar = FAR_PLANE;
		float t1, t2;
		int normal_face = 0; // 0: x, 1: y, 2: z

		// x-axis
		if(intersectAxis(objRay.origin()[0], objRay.direction()[0], -_width/2.f, _width/2.f, t1, t2)) {
			if(t1 > tnear)
				tnear = t1;
			if(t2 < tfar)
				tfar = t2;
		}
		// y-axis
		if(intersectAxis(objRay.origin()[1], objRay.direction()[1], -_height/2.f, _height/2.f, t1, t2)) {
			if(t1 > tnear) {
				tnear = t1;
				normal_face = 1;
			}
			if(t2 < tfar)
				tfar = t2;
		}
		// z-axis
		if(intersectAxis(objRay.origin()[2], objRay.direction()[2], -_depth/2.f, _depth/2.f, t1, t2)) {
			if(t1 > tnear) {
				tnear = t1;
				normal_face = 2;
			}
			if(t2 < tfar)
				tfar = t2;
		}
		// test for hit
		if((tnear > tfar) || (tfar <= 0))
			return false;

		tmin = tnear;

		// create the hit object
		h._position = objRay.pointAtParameter(tnear);

		// calculate normal
		if(normal_face == 0) {
			h._normal = SbVec3f(1.f, 0.f, 0.f);
			if(h._position[0] < 0)
				h._normal *= -1.f;
		}
		else if(normal_face == 1) {
			h._normal = SbVec3f(0.f, 1.f, 0.f);
			if(h._position[1] < 0)
				h._normal *= -1.f;
		}
		else { //if(normal_face == 2)
			h._normal = SbVec3f(0.f, 0.f, 1.f);
			if(h._position[2] < 0)
				h._normal *= -1.f;
		}

		// back to world space
		_transform.multVecMatrix(h._position, h._position);
		_transform.multDirMatrix(h._normal, h._normal);
		h._normal.normalize();
		// set object
		h.obj = (Object3d*)this;

		return true;
	}

	float intersectTrans(Ray &r, float &tmin, float &trans) {
		// transform ray into object space
		Ray objRay = Ray(r.direction(), r.origin());
		objRay.transform(_transInv);

		float tnear = -FAR_PLANE;
		float tfar = FAR_PLANE;
		float t1, t2;
		int normal_face = 0; // 0: x, 1: y, 2: z

		// x-axis
		if(intersectAxis(objRay.origin()[0], objRay.direction()[0], -_width/2.f, _width/2.f, t1, t2)) {
			if(t1 > tnear)
				tnear = t1;
			if(t2 < tfar)
				tfar = t2;
		}
		// y-axis
		if(intersectAxis(objRay.origin()[1], objRay.direction()[1], -_height/2.f, _height/2.f, t1, t2)) {
			if(t1 > tnear) {
				tnear = t1;
				normal_face = 1;
			}
			if(t2 < tfar)
				tfar = t2;
		}
		// z-axis
		if(intersectAxis(objRay.origin()[2], objRay.direction()[2], -_depth/2.f, _depth/2.f, t1, t2)) {
			if(t1 > tnear) {
				tnear = t1;
				normal_face = 2;
			}
			if(t2 < tfar)
				tfar = t2;
		}
		// test for hit
		if((tnear > tfar) || (tfar <= 0)) 
			return false;

		tmin = tnear;
		trans = _transparency;
		return true;
	}

private:
	float _width, _height, _depth;

	bool intersectAxis(float r, float d, float x1, float x2, float &t1, float &t2) {
		// check for parallel
		if(d == 0.f) {
			if((r < x1) || (r > x2))
				return false;
		}
		else { // not parallel
			t1 = (x1 - r)/d;
			t2 = (x2 - r)/d;
			if(t1 > t2) { // swap if the order is wrong
				float temp = t1;
				t1 = t2;
				t2 = temp;
			}
		}
		return true;
	}
};

class Light {
public:
	Light(const SbVec3f &location, const float intensity, const SbColor &color) {
		_location = location;
		_intensity = intensity;
		_color = color;

		// set up area light coordinate system (facing origin)
		SbVec3f w = SbVec3f(_location);
		w.normalize();
		SbVec3f t = SbVec3f(w);
		if((fabs(t[0]) <= fabs(t[1])) && (fabs(t[0]) <= fabs(t[2])))
			t[0] = 1.f;
		else if((fabs(t[1]) <= fabs(t[0])) && (fabs(t[1]) <= fabs(t[2])))
			t[1] = 1.f;
		else
			t[2] = 1.f;
		_u = t.cross(w);
		_u.normalize();
		_v = w.cross(_u);
	}

	// attributes
	SbVec3f _location;
	SbVec3f _u, _v; // used to make area lights
	float _intensity;
	SbColor _color;
};

class Scene {
public:
	Scene() {
		_objList = SbPList();
		_lightList = SbPList();
		_obj = _light = 0;
	}

	void addObj(Object3d* obj) {
		_objList.append((void*)obj);
	}

	void addLight(Light* light) {
		_lightList.append((void*)light);
	}

	Object3d*& getFirstObj() {
		_obj = 0;
		return (Object3d*&)_objList[0];	    
	}
	Object3d*& getNextObj() {
		_obj++;
		return (Object3d*&)_objList[_obj];	
	}

	Light*& getFirstLight() {
		_light = 0;
		return (Light*&)_lightList[0];	    
	}
	Light*& getNextLight() {
		_light++;
		return (Light*&)_lightList[_light];	
	}

private:
	int _obj, _light;
	SbPList _objList;
	SbPList _lightList;
};

class Camera {
public:
	Camera() {
		// set default camera orientation
		_eye = SbVec3f(0.f, 0.f, 10.f);
		_n = SbVec3f(0.f, 0.f, 1.f);
		_u = SbVec3f(1.f, 0.f, 0.f);
		_v = SbVec3f(0.f, 1.f, 0.f);
		_aspect = 1.f;
		_hAngle = 0.785398f;
	}

	void setPosition(const SbVec3f position) {
		_eye = position;
	}
	void setOrientation(const SbVec3f direction, const SbVec3f up) {
		// calculate n, v, u
		_n = -1.f*direction;
		_u = up.cross(_n);
		_v = _n.cross(_u);
	}
	SbVec3f getPosition () {
		return _eye;
	}
	void setAspect(const float aspect) {
		_aspect = aspect;
	}
	void setHeightAngle(const float hAngle) {
		_hAngle = hAngle;
	}
	void setResolution(const int resolution) {
		_yres = resolution;
	}
	void rayTraceSetup(Scene *s) {
		float _pixelHeight, _pixelWidth;
		SbColor color;

		// precalculate values
		_xres = (int)(_yres * _aspect);
		_pixelHeight = _pixelWidth = 2.f*tan(_hAngle/2.f) / _yres;

		_start = _eye + (-1.f*_n) - (_xres/2.f)*_pixelWidth*_u + (_yres/2.f)*_pixelHeight*_v + (_pixelWidth/2.f)*_u - (_pixelHeight/2.f)*_v;

		_pixelU = _pixelWidth*_u;
		_pixelV = _pixelHeight*_v;
		// subdivision of a pixel
		/*if(RECONST == 0)
		{
			// 1/8 of a pixel for box filter
			_pixelU_sub = (_pixelWidth/8)*_u;
			_pixelV_sub = (_pixelHeight/8)*_v;
		}
		else
		{
			// 1/4 of a pixel for tent and m-n filter
			_pixelU_sub = (_pixelWidth/4)*_u;
			_pixelV_sub = (_pixelHeight/4)*_v;
		}*/

	    srand(time(NULL)); // seed random numbers

		ray_count = 0;
		aa_count = 0;

		//////////////////////////////////////
		// setup array for adaptive anti-aliasing
		_pixels = new float[(_xres*2+6)*3];
		_pixels_pos = 0;

		for(int i=0; i < _xres*2+6; i++) {
			_pixels[pixels_offset(i) + 0] = -1.f;
		}

		for(int p=0; p<_xres; p++) {
			color = rayTracePixel(s, 0, p, false, false);

			_pixels[pixels_offset(p) + 0] = color[0];
			_pixels[pixels_offset(p) + 1] = color[1];
			_pixels[pixels_offset(p) + 2] = color[2];
		}

		color = rayTracePixel(s, 1, 0, false, false);
		_pixels[pixels_offset(_xres+2) + 0] = color[0];
		_pixels[pixels_offset(_xres+2) + 1] = color[1];
		_pixels[pixels_offset(_xres+2) + 2] = color[2];

	}

	void rayTraceScanline(Scene *s, Bitmap *canvas, int sl, bool quick, bool aa) {
		int r, g, b;
		SbColor color = SbColor(0.f, 0.f, 0.f);
		SbColor new_color = SbColor(-1.f, 0.f, 0.f);

		for(int p=0; p<_xres; p++) {
			if((quick) || (!aa)) {
				color = rayTracePixel(s, sl, p, quick, false);
			}
			else {
				color[0] = _pixels[pixels_offset(_pixels_pos) + 0];
				color[1] = _pixels[pixels_offset(_pixels_pos) + 1];
				color[2] = _pixels[pixels_offset(_pixels_pos) + 2];
				if((sl < _yres-2) || ((sl < _yres-1) && (p < _xres-1))) {
					if(p < _xres-1) {
						new_color = rayTracePixel(s, sl+1, p+1, false, false);
						if(needAA(color, new_color, sl, p)) {
							color = rayTracePixel(s, sl, p, false, true);
							aa_count++;
						}

						// shift array
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 2] = new_color[2];
						_pixels_pos = pos_offset(_pixels_pos+1);
					}
					else {
						if(needAA(color, new_color, sl, p)) {
							color = rayTracePixel(s, sl, p, false, true);
							aa_count++;
						}

						// shift array
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 2] = new_color[2];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 2] = new_color[2];
						new_color = rayTracePixel(s, sl+2, 0, false, false);
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 2] = new_color[2];
						_pixels_pos = pos_offset(_pixels_pos+3);
					}
				}
				else {
					if(p < _xres-1) {
						// last row
						if(needAA(color, new_color, sl, p)) {
							color = rayTracePixel(s, sl, p, false, true);
							aa_count++;
						}

						// shift array
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 2] = new_color[2];
						_pixels_pos = pos_offset(_pixels_pos+1);
					}
					else {
						if(needAA(color, new_color, sl, p)) {
							color = rayTracePixel(s, sl, p, false, true);
							aa_count++;
						}

						// shift array
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+3) + 2] = new_color[2];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+4) + 2] = new_color[2];
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 0] = new_color[0];
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 1] = new_color[1];
						_pixels[pixels_offset(_pixels_pos+_xres+5) + 2] = new_color[2];
						_pixels_pos = pos_offset(_pixels_pos+3);
					}
				}
			}
			// clamp
			if(color[0] > 1.0f)
				color[0] = 1.0f;
			if(color[0] < 0.0f)
				color[0] = 0.0f;
			if(color[1] > 1.0f)
				color[1] = 1.0f;
			if(color[1] < 0.0f)
				color[1] = 0.0f;
			if(color[2] > 1.0f)
				color[2] = 1.0f;
			if(color[2] < 0.0f)
				color[2] = 0.0f;
			r = (int)(color[0]*255.f);
			g = (int)(color[1]*255.f);
			b = (int)(color[2]*255.f);
			canvas->SetPixel(p, sl, Color::FromArgb(r, g, b));
		}
	}

	unsigned int ray_count, aa_count;

private:
	SbVec3f _eye;
	SbVec3f _n, _v, _u;
	float _aspect, _hAngle;
	int _yres;

	// calculated
	int _xres;
	SbVec3f _pixelU, _pixelV, _start; //_pixelU8, _pixelV8, 
	float *_pixels;
	int _pixels_pos;

	int pixels_offset(int pos) {
		return pos_offset(pos)*3;
	}

	int pos_offset(int pos) {
		while(pos >= _xres*2+6)
			pos -= (_xres*2+6);
		while(pos < 0)
			pos += _xres*2+6;
		return pos;
	}

	SbColor rayTracePixel(Scene *s, int sl, int p, bool quick, bool aa) {
		Ray ray;
		SbColor color = SbColor(0.f, 0.f, 0.f);
		SbVec3f pixel, rayDir, subpixel;
			
		pixel = _start - _pixelV*sl + _pixelU*p;

		if((!aa) || (quick)) {
			rayDir = pixel - _eye; // get ray direction
			rayDir.normalize();
			ray = Ray(rayDir, _eye); // create ray
			color = shade(ray, s, 0, 0, true, quick);
		}
		else {
			float scale = (RECONST == 0) ? 2.0f : 1.0f;
			float sum = 0.0f;
			float weight;

			for(int subsl=-3; subsl<=3; subsl+=2) {
				for(int subp=-3; subp<=3; subp+=2) {
					float U_offset, V_offset;

					// subsample
					if((SAMPLING == 0) || (SAMPLING == 1))
					{
						// uniform grid
						U_offset = ((float)subp)/(4.0f * scale);
						V_offset = ((float)subsl)/(4.0f * scale);

						if(SAMPLING == 1)
						{
							// jittering
							U_offset += randFloat()/(4.0f * scale);
							V_offset += randFloat()/(4.0f * scale);
							//subpixel += _pixelU_sub * randU; // jitter U
							//subpixel += _pixelV_sub * randV; // jitter V
						}
					}
					else //if(SAMPLING == 2)
					{
						U_offset = randFloat()/scale;
						V_offset = randFloat()/scale;
					}
					
					subpixel = pixel + U_offset*_pixelU + V_offset*_pixelV;
					rayDir = subpixel - _eye; // get ray direction
					rayDir.normalize();
					ray = Ray(rayDir, _eye); // create ray

					if(RECONST == 0) // box
					{
						weight = 1.0f;
					}
					else if(RECONST == 1) // tent
					{
						weight = (1 - sqrt(U_offset*U_offset*(0.7071) + V_offset*V_offset*(0.7071)));
					}
					else if(RECONST == 2) // m-n
					{
						float dist = sqrt(U_offset*U_offset + V_offset*V_offset) * 2.0f;
						if(dist < 1.0f)
							weight = 7*dist*dist*dist - 12*dist*dist + 16/3;
						else if(dist < 2.0f)
							weight = -7/3*dist*dist*dist + 12*dist*dist - 20*dist + 32/3;
						else
							weight = 0.0f;
					}
					sum += weight;
					color += shade(ray, s, 0, 0, true, false) * weight;
				}
			}
			// average
			color /= sum;
		}
		return color;
	}

	bool needAA(SbColor &color, SbColor &new_color, int sl, int p) {
		// test for large differences in color
		// The Row Above
		/*if(_pixels[pixels_offset(_pixels_pos-_xres-1) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos-_xres-1) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos-_xres-1) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos-_xres-1) + 2]) > AA_ALPHA) return true;
		}*/
		if(_pixels[pixels_offset(_pixels_pos-_xres) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos-_xres) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos-_xres) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos-_xres) + 2]) > AA_ALPHA) return true;
		}
		/*if(_pixels[pixels_offset(_pixels_pos-_xres+1) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos-_xres+1) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos-_xres+1) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos-_xres+1) + 2]) > AA_ALPHA) return true;
		}*/
		// The Sides
		if(_pixels[pixels_offset(_pixels_pos-1) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos-1) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos-1) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos-1) + 2]) > AA_ALPHA) return true;
		}
		if(_pixels[pixels_offset(_pixels_pos+1) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos+1) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos+1) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos+1) + 2]) > AA_ALPHA) return true;
		}
		// The Row Below
		/*if(_pixels[pixels_offset(_pixels_pos+_xres-1) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos+_xres-1) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos+_xres-1) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos+_xres-1) + 2]) > AA_ALPHA) return true;
		}*/
		if(_pixels[pixels_offset(_pixels_pos+_xres) + 0] != -1) {
			if(fabs(color[0] - _pixels[pixels_offset(_pixels_pos+_xres) + 0]) > AA_ALPHA) return true;
			if(fabs(color[1] - _pixels[pixels_offset(_pixels_pos+_xres) + 1]) > AA_ALPHA) return true;
			if(fabs(color[2] - _pixels[pixels_offset(_pixels_pos+_xres) + 2]) > AA_ALPHA) return true;
		}
		/*if(new_color[0] != -1) {
			if(fabs(color[0] - new_color[0]) > AA_ALPHA) return true;
			if(fabs(color[1] - new_color[1]) > AA_ALPHA) return true;
			if(fabs(color[2] - new_color[2]) > AA_ALPHA) return true;
		}*/
		return false;
	}

	bool objectIntersect(Ray &r, Scene *s, Hit &hit, float &tmin) {
		float t;
		Hit	h;
		bool hitObj	= false;

		ray_count++;

		Object3d *obj =	s->getFirstObj();
		while(obj != NULL) {
			if(obj->intersect(r, h,	t))	{ // detect	intersection
				if(t < tmin) { // check	if closest
					tmin = t;
					hit	= h;
					hitObj = true;
				}
			}
			obj	= s->getNextObj(); // go	to next	object
		}

		return hitObj;
	}

	float objectShadowIntersect(Ray &r, Scene *s, float tmin) {
		float t;
		float obj_trans;
		float trans = 1.0f;

		ray_count++;

		Object3d *obj = s->getFirstObj();
		while(obj != NULL) {
			// get transparency of hit object
			if(obj->intersectTrans(r, t, obj_trans)) {
				if(t < tmin) { // check if valid
					trans *= obj_trans; // add transparency
					if(trans == 0.f)
						return trans;
				}
			}
			obj = s->getNextObj(); // go to next object
		}

		return trans;
	}

	SbColor texture0(SbVec3f &pos, Object3d* obj) {
		SbVec3f obj_pos;
		// transform position to object space
		obj->_transInv.multVecMatrix(pos, obj_pos);

		int dif = (int)(obj_pos[0]*5+2000)%2;
		if(dif == 0)
			return SbColor(1.f, 0.f, 0.f);
		return SbColor(0.f, 1.f, 0.f);
	}

	SbColor texture1(SbVec3f &pos, Object3d* obj) {
		SbVec3f obj_pos;
		// transform position to object space
		obj->_transInv.multVecMatrix(pos, obj_pos);

		int dif = ((int)(obj_pos[0]*5+2000) + (int)(obj_pos[1]*5+2000))%2;
		if(dif == 0)
			return SbColor(0.2f, 0.2f, 0.2f);
		return SbColor(0.9f, 0.9f, 0.9f);
	}

	SbColor texture2(SbVec3f &pos, Object3d* obj) {
		SbVec3f obj_pos;
		// transform position to object space
		obj->_transInv.multVecMatrix(pos, obj_pos);

		int dif = ((int)(sqrt(26*obj_pos[0]*obj_pos[0]+26*obj_pos[1]*obj_pos[1])))%2;
		if(dif == 0)
			return SbColor(1.f, 0.f, 0.f);
		return SbColor(1.f, 1.f, 1.f);
	}

	SbColor texture3(SbVec3f &pos, Object3d* obj) {
		SbVec3f obj_pos, norm;
		// transform position to object space
		obj->_transInv.multVecMatrix(pos, obj_pos);
		norm = SbVec3f(obj_pos[0], obj_pos[1], 0.f);
		norm.normalize();
		float dot_prod = cos(acos(norm.dot(SbVec3f(1.f, 0.f, 0.f)))*5)/4;

		int dif = ((int)((sqrt(26*obj_pos[0]*obj_pos[0]+26*obj_pos[1]*obj_pos[1]))+dot_prod))%2;
		if(dif == 0)
			return SbColor(1.f, 0.f, 1.f);
		return SbColor(1.f, 1.f, 0.f);
	}

	SbColor texture4(SbVec3f &pos, Object3d* obj) {
		SbVec3f obj_pos;
		// transform position to object space
		obj->_transInv.multVecMatrix(pos, obj_pos);
		float r = cos(((obj_pos[0]-obj_pos[1])*4)+1.5708)/2 + 0.5;
		float g = cos((obj_pos[0]-obj_pos[1])*4)/2 + 0.5;
		float b = cos(((obj_pos[0]-obj_pos[1])*4)-1.5708)/2 + 0.5;

		return SbColor(r, g, b);
	}

	SbColor illuminate(SbVec3f &v, SbVec3f &n, SbVec3f &pos, Scene *s, Object3d* obj, bool quick) {
		SbColor c, dif_tex;
		SbVec3f l, r, soft_l;
		float dif_comp, q, spec_comp, light_t, vr, intens;
		float trans;
		Ray ray;
		Hit h;
		int texture;

		// add ambient component
		c = 0.2f*obj->_ambient;

		Light *light = s->getFirstLight();
		while(light != NULL) {
			// calculate light vector
			l = light->_location - pos;
			light_t = l.length() - EPSILON;
			l.normalize();
			dif_comp = n.dot(l);
			// check for front face
			if(dif_comp > 0.f) {
				if(!quick) {
					// check for shadow
					//ray = Ray(l, pos + EPSILON*l); // shoot ray to light
					// step out epsilon
					// get transparency of shadow
					trans = 0.f;//objectShadowIntersect(ray, s, light_t);

					// soft shadows

					// shoot 4 rays to see if it is necessary to do soft shadows
					for(int u=-1; u<=1; u+=2) {
						for(int v=-1; v<=1; v+=2) {
							soft_l = light->_location - pos;
							soft_l += light->_u*u + light->_v*v;
							soft_l += light->_u*randFloat()*0.25f + light->_v*randFloat()*0.25f;
							light_t = soft_l.length() - EPSILON;
							soft_l.normalize();
							ray = Ray(soft_l, pos + EPSILON*soft_l);
							trans += objectShadowIntersect(ray, s, light_t);
						}
					}
					// if full shadow than skip
					if(trans <= 0.f) {
						intens = 0.f;
					}
					else {
						// if shadows were detected, then shoot more rays
						if(trans/4.f < 1.f) {
							for(float u=-1; u<=1; u+=0.5) {
								for(float v=-1; v<=1; v+=0.5) {
									// don't repeat rays already used
									if(!(((u==-1)||(u==1)) && ((v==-1)||(v==1)))) {
										soft_l = light->_location - pos;
										soft_l += light->_u*u + light->_v*v;
										soft_l += light->_u*randFloat()*0.25f + light->_v*randFloat()*0.25f;
										light_t = soft_l.length() - EPSILON;
										soft_l.normalize();
										ray = Ray(soft_l, pos + EPSILON*soft_l);
										trans += objectShadowIntersect(ray, s, light_t);
									}
								}
							}

							// avg shadows
							intens = light->_intensity * trans/25.f;
						}
						else {
							intens = light->_intensity;
						}
					}
				}
				else {
					intens = light->_intensity;
				}

				// add diffuse component
				if(obj->_texture == 0)		    
					c += dif_comp * obj->_diffuse * intens; // no-texture
				else {
					texture = obj->_texture%5;
					if(texture == 0)
						dif_tex = texture1(pos, obj);
					if(texture == 1)
						dif_tex = texture4(pos, obj);
					if(texture == 2)
						dif_tex = texture2(pos, obj);
					if(texture == 3)
						dif_tex = texture3(pos, obj);
					if(texture == 4)
						dif_tex = texture0(pos, obj);
					c += dif_comp * dif_tex * intens;
				}

				// add specular
				r = 2.f*dif_comp*n - l; // reflectance vector
				q = obj->_shininess;
				q = q*q*1000.f; // adjust q
				vr = v.dot(r);
				if((vr >= 0.f) && (q > 0.f)) { // check for 90 degrees
					// with edge specs
					spec_comp = pow(vr, (int)q) * intens;
					if(spec_comp != spec_comp) {
						spec_comp = 0.f;
					}
					c[0] +=  spec_comp * obj->_specular[0] * light->_color[0];
					c[1] +=  spec_comp * obj->_specular[1] * light->_color[1];
					c[2] +=  spec_comp * obj->_specular[2] * light->_color[2];
				}		
			}

			light = s->getNextLight();
		}

		return c;
	}

	SbColor shade(Ray &r, Scene *s, int reflrec, int refrrec, bool outside, bool quick) {
		Hit closest;
		SbColor c = SbColor(0.f, 0.f, 0.f);

		float tmin = FAR_PLANE; // set minimum t to infinity (FAR_PLANE)

		// check for an intersection
		if(objectIntersect(r, s, closest, tmin)) {
			// calculate simple illumination
			Object3d *obj = closest.obj;
			SbVec3f v = r.direction()*-1.f; // view vector
			SbVec3f n = closest._normal; // normal
			SbVec3f pos = closest._position; // position

			// flip normal on inside
			if(!outside)
				n *= -1.f;
			float nv = n.dot(v);

			// Simple illumination
			if(outside)
				c += illuminate(v, n, pos, s, obj, quick);

			if(!quick) {
				// Reflection
				float shine = obj->_shininess;
				if((shine > 0.01) && (reflrec < MAXREFLRECURSION)) {
					SbVec3f refl_vec = 2.f*nv*n - v;
					Ray refl_ray = Ray(refl_vec, pos + EPSILON*refl_vec);
					SbColor refl_col = shade(refl_ray, s, ++reflrec, refrrec, outside, false);
					c += refl_col * shine;
				}

				// Refraction
				float trans = obj->_transparency;
				if((trans > 0.0) && (refrrec < MAXREFRRECURSION)) {
					if(outside || (nv > 0.48038446f)) // check for total internal ref.
					{
						//float nr = outside ? 1.5f : 1.f/1.5f;
						float nr = outside ? 1.f/1.75f : 1.75f; // 1.3 looks
						// better imho.
						float coef = (nr*nv - sqrt(1.f-nr*nr*(1.f-nv*nv)));

						SbVec3f refr_vec = coef*n - nr*v;
						Ray refr_ray = Ray(refr_vec, pos + EPSILON*refr_vec);
						SbColor refr_col = shade(refr_ray, s, reflrec, ++refrrec, !outside, false);
						c += refr_col * trans;
					}
				}
			}

			// clamp
			if(c[0] > 1.f)
				c[0] = 1.f;
			if(c[1] > 1.f)
				c[1] = 1.f;
			if(c[2] > 1.f)
				c[2] = 1.f;

			return c;
		}

		return SbColor(0.f, 0.f, 0.f); // no hit, return black
	}
};


int raytrace_load(char *input, int rows, Scene *_scene, Camera *_camera, bool textures) 
{
	int num = 1;

	// my objects
	//Scene _scene = Scene(); // scene to be raytraced
	//Camera _camera = Camera();

	// Inventor setup
	SoDB::init();
	OSUInventorScene *scene = new OSUInventorScene(input);

	//cout << "Read objects from file " << argv[1] << "." << endl;
	//cout << "Number of objects = " 
	// << scene->Objects.getLength() << "." << endl;

	// list objects
	for (int i = 0; i < scene->Objects.getLength(); i++) {
		OSUObjectData * obj = (OSUObjectData *)scene->Objects[i];

		if (!(obj->Check())) {
			//cerr << "Error detected in OSUObjectData for object " << i << "." 
			// << endl;
			exit(20);
		}

		SoType shape_type = obj->shape->getTypeId();

		//cout << "Object " << i << " is a "
		//     << shape_type.getName().getString() << "." << endl;

		if (shape_type == SoSphere::getClassTypeId()) {
			SoSphere * sphere = (SoSphere *) obj->shape;
			// cout << "  Sphere radius = " << sphere->radius.getValue() << "."
			// << endl;
		}

		if (shape_type == SoCube::getClassTypeId()) {
			SoCube * cube = (SoCube *) obj->shape;
			// cout << "  Cube width, height, depth = (" 
			// << cube->width.getValue() << ","
			// << cube->height.getValue() << ","
			// << cube->depth.getValue() << ")." << endl;
		}

		// object transformation
		SoTransform * transformation = obj->transformation;
		SbVec3f scale_vector = transformation->scaleFactor.getValue();
		SbRotation rotation = transformation->rotation.getValue();
		SbVec3f rotation_axis;
		float rotation_angle;
		rotation.getValue(rotation_axis, rotation_angle);
		SbVec3f translation_vector = transformation->translation.getValue();
		//cout << "  Scale by (" 
		//    << scale_vector[0] << ","
		//     << scale_vector[1] << ","
		//     << scale_vector[2] << ")." << endl;
		//cout << "  Rotate around axis (" 
		//     << rotation_axis[0] << ","
		//     << rotation_axis[1] << ","
		//     << rotation_axis[2] << ")" 
		//     << " by " << rotation_angle << " radians." << endl;
		//cout << "  Translate by (" 
		//     << translation_vector[0] << ","
		//     << translation_vector[1] << ","
		//     << translation_vector[2] << ")." << endl;

		// object material (color)
		SoMaterial * material = obj->material;
		// each material property is an array
		// use only the first entry in that array
		//cout << "  Diffuse color (red, green, blue) = ("
		//     << material->diffuseColor[0][0] << ","
		//     << material->diffuseColor[0][1] << ","
		//     << material->diffuseColor[0][2] << ")." << endl;
		//cout << "  Specular color (red, green, blue) = ("
		//     << material->specularColor[0][0] << ","
		//     << material->specularColor[0][1] << ","
		//     << material->specularColor[0][2] << ")." << endl;
		//cout << "  Ambient color (red, green, blue) = ("
		//     << material->ambientColor[0][0] << ","
		//     << material->ambientColor[0][1] << ","
		//     << material->ambientColor[0][2] << ")." << endl;
		//cout << "  Shininess (specular exponent) = " 
		//     << material->shininess[0] << "." << endl;
		//cout << "  Transparency (specular transmissive coefficient) = " 
		//     << material->transparency[0] << "." << endl;


		////////////////////////////////////////
		// add object to raytracer scene
		if (shape_type == SoSphere::getClassTypeId()) {
			SoSphere * sphere = (SoSphere *) obj->shape;

			Sphere *_sphere = new Sphere(sphere->radius.getValue(), num++);
			_sphere->setTransform(translation_vector, rotation, scale_vector);
			_sphere->setAmbient(material->ambientColor[0]);
			_sphere->setDiffuse(material->diffuseColor[0]);
			_sphere->setSpecular(material->specularColor[0]);
			_sphere->setShininess(material->shininess[0]);
			_sphere->setTransparency(material->transparency[0]);
			if((material->transparency[0] > 0.0f) || (!textures))
				_sphere->_texture = 0;

			_scene->addObj(_sphere);
		};

		if (shape_type == SoCube::getClassTypeId()) {
			SoCube * cube = (SoCube *) obj->shape;

			Cube *_cube = new Cube(cube->width.getValue(), cube->height.getValue(), cube->depth.getValue(), num++);
			_cube->setTransform(translation_vector, rotation, scale_vector);
			_cube->setAmbient(material->ambientColor[0]);
			_cube->setDiffuse(material->diffuseColor[0]);
			_cube->setSpecular(material->specularColor[0]);
			_cube->setShininess(material->shininess[0]);
			_cube->setTransparency(material->transparency[0]);
			if((material->transparency[0] > 0.0f) || (!textures))
				_cube->_texture = 0;

			_scene->addObj(_cube);
		};
	};

	// list lights
	//cout << endl;
	for (int j = 0; j < scene->Lights.getLength(); j++) {
		SoLight * light = (SoLight *) scene->Lights[j];
		SoType light_type = light->getTypeId();

		if (light->on.getValue() == true){
			//cout << "Light " << j << " is on." << endl;
		}
		else{
			//cout << "Light " << j << " is off." << endl;
		}

		float light_intensity = light->intensity.getValue();
		SbColor light_color = light->color.getValue();

		//cout << "  Light intensity = " << light_intensity << endl;
		//cout << "  Light color = (" 
		//     << light_color[0] << ","
		//     << light_color[1] << ","
		//     << light_color[2] << ")." << endl;


		if (light_type == SoPointLight::getClassTypeId()) {
			SoPointLight * point_light = (SoPointLight *) light;
			SbVec3f location = point_light->location.getValue();
			//cout << "  Point light location = (" 
			// << location[0] << ","
			// << location[1] << ","
			// << location[2] << ")." << endl;

			// Add the light to the scene if it is on.
			if (light->on.getValue() == true) {
				Light *_light = new Light(location, light_intensity, light_color);
				_scene->addLight(_light);
			}

		};
	};

	// camera
	if (scene->Camera == NULL) {
		//cout << endl;
		//cout << "No camera found." << endl;
	}
	else {
		SoCamera * camera = scene->Camera;
		SbVec3f camera_position = camera->position.getValue();
		SbRotation camera_orientation = camera->orientation.getValue();
		SbVec3f camera_rotation_axis;
		float camera_rotation_angle;
		camera_orientation.getValue(camera_rotation_axis, camera_rotation_angle);
		float camera_aspect_ratio = camera->aspectRatio.getValue();
		SoType camera_type = camera->getTypeId();

		// calculate camera direction and camera up direction
		SbVec3f camera_direction, camera_up;
		camera_orientation.multVec(SbVec3f(0, 0, -1), camera_direction);
		camera_orientation.multVec(SbVec3f(0, 1, 0), camera_up);

		//cout << endl;
		//cout << "Camera position = ("
		//     << camera_position[0] << ","
		//     << camera_position[1] << ","
		//     << camera_position[2] << ")." << endl;
		//cout << "Camera rotation axis = (" 
		//     << camera_rotation_axis[0] << ","
		//     << camera_rotation_axis[1] << ","
		//     << camera_rotation_axis[2] << ")." << endl;
		//cout << "Camera rotation angle = " << camera_rotation_angle << " radians."
		//     << endl;
		//cout << "Camera direction = (" 
		//     << camera_direction[0] << ","
		//     << camera_direction[1] << ","
		//     << camera_direction[2] << ")." << endl;
		//cout << "Camera up direction = ("
		//     << camera_up[0] << ","
		//     << camera_up[1] << ","
		//     << camera_up[2] << ")." << endl;
		//cout << "Camera width/height aspect ratio = "
		//     << camera_aspect_ratio << "." << endl;


		if (camera_type == SoPerspectiveCamera::getClassTypeId()) {
			// camera is a perspective camera
			SoPerspectiveCamera * perspective_camera = 
				(SoPerspectiveCamera *) camera;
			float camera_height_angle = perspective_camera->heightAngle.getValue();
			//cout << "Perspective camera height angle = "
			// << camera_height_angle << " radians." << endl;

			//////////////////////
			// setup raytracer camera
			_camera->setPosition(camera_position);
			_camera->setOrientation(camera_direction, camera_up);
		};
	};

	_camera->setResolution(rows);
	//cout << "\nImage file height " << argv[3] << ".\n";
	//cout << "Writing image to file " << argv[2] << "...\n";

	//_camera.rayTraceScene(_scene, canvas);

	return(1);

}
